import csv
import json
import isodate
from pathlib import Path

from dataflows.helpers.extended_json import (
    DATETIME_F_FORMAT, DATE_F_FORMAT, TIME_F_FORMAT,
    DATETIME_P_FORMAT, DATE_P_FORMAT, TIME_P_FORMAT,
)

from .base import FileFormat, json_dumps


class CsvTitlesDictWriter(csv.DictWriter):
    def __init__(self, *args, **kwargs):
        self.fieldtitles = kwargs.pop('fieldtitles')
        super().__init__(*args, **kwargs)

    def writeheader(self):
        header = dict(zip(self.fieldnames, self.fieldtitles))
        self.writerow(header)


class CSVFormat(FileFormat):

    SERIALIZERS = {
        'array': json_dumps,
        'object': json_dumps,
        'datetime': lambda d: d.strftime(DATETIME_F_FORMAT),
        'date': lambda d: d.strftime(DATE_F_FORMAT),
        'time': lambda d: d.strftime(TIME_F_FORMAT),
        'duration': lambda d: isodate.duration_isoformat(d),
        'geopoint': lambda d: '{}, {}'.format(*d),
        'geojson': json.dumps,
        'year': lambda d: '{:04d}'.format(d),
        'yearmonth': lambda d: '{:04d}-{:02d}'.format(*d),
    }
    NULL_VALUE = ''

    PYTHON_DIALECT = {
        'number': {
            'decimalChar': '.',
            'groupChar': ''
        },
        'date': {
            'format': DATE_P_FORMAT
        },
        'time': {
            'format': TIME_P_FORMAT
        },
        'datetime': {
            'format': DATETIME_P_FORMAT
        },
        'boolean': {
            'trueValues': ['True'],
            'falseValues': ['False']
        }
    }

    def __init__(self, file, schema, use_titles=False, **options):
        headers = [f.name for f in schema.fields]
        if use_titles:
            titles = [f.descriptor.get('title', f.name) for f in schema.fields]
            csv_writer = CsvTitlesDictWriter(file, headers, fieldtitles=titles)
        else:
            csv_writer = csv.DictWriter(file, headers)
        csv_writer.writeheader()
        super(CSVFormat, self).__init__(csv_writer, schema, **options)

    @classmethod
    def prepare_resource(cls, resource):
        descriptor = resource.descriptor
        descriptor['encoding'] = 'utf-8'
        descriptor['path'] = str(Path(descriptor['path']).with_suffix('.csv'))
        descriptor['format'] = 'csv'
        descriptor['dialect'] = dict(
            lineTerminator='\r\n',
            delimiter=',',
            doubleQuote=True,
            quoteChar='"',
            skipInitialSpace=False
        )
        super(CSVFormat, cls).prepare_resource(resource)

    def write_transformed_row(self, transformed_row):
        self.writer.writerow(transformed_row)

    def finalize_file(self):
        pass



import sys

from datapackage import Package

from ..helpers.extended_json import ejson


def unstream(file=sys.stdin):

    if isinstance(file, str):
        file = open(file)

    def read():
        line = file.readline().strip()
        if len(line) > 0:
            return ejson.loads(line)
        return None

    def res_reader():
        while True:
            r = read()
            if r is not None:
                yield r
            else:
                break

    def func(package):
        descriptor = read()
        yield Package(descriptor)
        for _ in descriptor['resources']:
            yield res_reader()

    return func



from .. import DataStreamProcessor


class conditional(DataStreamProcessor):

    def __init__(self, predicate, flow):
        super().__init__()
        self.predicate = predicate
        self.flow = flow

    def _process(self):
        ds = self.source._process()
        if self.predicate(ds.dp):
            if callable(self.flow):
                flow = self.flow(ds.dp)
            else:
                flow = self.flow
            return flow.datastream(ds)
        else:
            return ds



import datetime
import decimal
import os
import logging
import copy
import json

from tableschema_sql import Storage
from sqlalchemy import create_engine

from tableschema.exceptions import ValidationError

from ...base import ResourceWrapper

from .dumper_base import DumperBase


def jsonize(obj):
    return json.dumps(obj)


def strize(obj):
    if isinstance(obj, dict):
        return dict(
            (k, strize(v))
            for k, v in obj.items()
        )
    elif isinstance(obj, (str, int, float, bool)):
        return obj
    elif isinstance(obj, datetime.date):
        return obj.isoformat()
    elif isinstance(obj, decimal.Decimal):
        return float(obj)
    elif isinstance(obj, (list, set)):
        return [strize(x) for x in obj]
    elif obj is None:
        return None
    assert False, "Don't know how to handle object %r" % obj


OBJECT_FIXERS = {
    'sqlite': [strize, jsonize],
    'postgresql': [strize]
}


class SQLDumper(DumperBase):

    def __init__(self,
                 tables,
                 engine='env://DATAFLOWS_DB_ENGINE',
                 updated_column=None, updated_id_column=None,
                 **options):
        super(SQLDumper, self).__init__(options)
        table_to_resource = tables

        if isinstance(engine, str):
            if engine.startswith('env://'):
                env_var = engine[6:]
                engine = os.environ.get(env_var)
                if engine is None:
                    raise ValueError("Couldn't connect to DB - "
                                     "Please set your '%s' environment variable" % env_var)

            self.engine = create_engine(engine)
            # check connection
            with self.engine.connect():
                pass
        else:
            self.engine = engine

        for k, v in table_to_resource.items():
            v['table-name'] = k

        self.converted_resources = \
            dict((v['resource-name'], v) for v in table_to_resource.values())

        self.updated_column = updated_column
        self.updated_id_column = updated_id_column
        self.batch_size = options.get('batch_size', 1000)
        self.use_bloom_filter = options.get('use_bloom_filter', True)

    def normalize_for_engine(self, dialect, resource, schema_descriptor):
        actions = {}
        for field in schema_descriptor['fields']:
            if field['type'] in ['array', 'object']:
                assert dialect in OBJECT_FIXERS, "Don't know how to handle %r connection dialect" % dialect
                actions.setdefault(field['name'], []).extend(OBJECT_FIXERS[dialect])

        for row in resource:
            for name, action_list in actions.items():
                for action in action_list:
                    row[name] = action(row.get(name))

            yield row

    def process_resource(self, resource: ResourceWrapper):
        resource_name = resource.res.name
        if resource_name not in self.converted_resources:
            return resource
        else:
            converted_resource = self.converted_resources[resource_name]
            mode = converted_resource.get('mode', 'rewrite')
            table_name = converted_resource['table-name']
            indexes_fields = converted_resource.get('indexes_fields', None)
            storage = Storage(self.engine, prefix=table_name)
            if mode == 'rewrite' and '' in storage.buckets:
                storage.delete('')
            schema_descriptor = resource.res.descriptor['schema']
            schema = self.normalize_schema_for_engine(self.engine.dialect.name,
                                                      schema_descriptor)
            if '' not in storage.buckets:
                logging.info('Creating DB table %s', table_name)
                try:
                    storage.create('', schema, indexes_fields=indexes_fields)
                except ValidationError as e:
                    logging.error('Error validating schema %r', schema_descriptor)
                    for err in e.errors:
                        logging.error('Error validating schema: %s', err)
                    raise
            else:
                storage.describe('', schema)

            update_keys = None
            if mode == 'update':
                update_keys = converted_resource.get('update_keys')
                if update_keys is None:
                    update_keys = schema_descriptor.get('primaryKey', [])
            logging.info('Writing to DB %s -> %s (mode=%s, keys=%s)',
                         resource_name, table_name, mode, update_keys)
            return map(self.get_output_row,
                       storage.write(
                           '',
                           self.normalize_for_engine(self.engine.dialect.name,
                                                     resource, schema_descriptor),
                           keyed=True, as_generator=True,
                           update_keys=update_keys,
                           buffer_size=self.batch_size,
                           use_bloom_filter=self.use_bloom_filter,
                       ))

    def get_output_row(self, written):
        row, updated, updated_id = written.row, written.updated, written.updated_id
        if self.updated_column:
            row[self.updated_column] = updated
        if self.updated_id_column:
            row[self.updated_id_column] = updated_id
        return row

    def normalize_schema_for_engine(self, dialect, schema):
        schema = copy.deepcopy(schema)
        for field in schema['fields']:
            if dialect == 'sqlite' and field['type'] in ['object', 'array']:
                field['type'] = 'string'
        return schema



import re

from .. import ResourceWrapper
from ..helpers.resource_matcher import ResourceMatcher


def process_resource(rows: ResourceWrapper, configuration):
    fields = configuration[rows.res.descriptor['name']]
    for row in rows:
        row = dict(
            (k, v)
            for k, v in row.items()
            if k in fields
        )
        yield row


def select_fields(fields, resources=None, regex=True):

    def func(package):
        configuration = dict()
        matcher = ResourceMatcher(resources, package.pkg)
        dp_resources = package.pkg.descriptor.get('resources', [])
        for resource in dp_resources:
            if matcher.match(resource['name']):
                configuration.setdefault(resource['name'], set())
                dp_fields = resource['schema'].get('fields', [])
                dp_fields = dict(
                    (f['name'], f)
                    for f in dp_fields
                )
                new_fields = []
                for selected_field in fields:
                    selected_field = re.compile('^{}$'.format(
                        selected_field
                        if regex
                        else re.escape(selected_field)))
                    for name in list(dp_fields.keys()):
                        if selected_field.match(name):
                            new_fields.append(dp_fields.pop(name))
                            configuration[resource['name']].add(name)

                assert len(new_fields) > 0, \
                    "Can't find any fields to select in resource %s" % resource['name']

                resource['schema']['fields'] = new_fields
        yield package.pkg

        for resource in package:
            if not matcher.match(resource.res.name):
                yield resource
            else:
                yield process_resource(resource, configuration)

    return func



import re
import copy
import os
import warnings
import collections

from kvfile import KVFile

from dataflows import PackageWrapper


# DB Helper
class KeyCalc(object):

    def __init__(self, key_spec):
        if isinstance(key_spec, list):
            key_list = key_spec
            key_spec = ':'.join('{%s}' % key for key in key_spec)
        else:
            key_list = re.findall(r'\{(.*?)\}', key_spec)
        self.key_spec = key_spec
        self.key_list = key_list

    def __call__(self, row, row_number):
        return self.key_spec.format(**{**row, '#': row_number})


# Aggregator helpers
def identity(x):
    return x


def median(values):
    if values is None:
        return None
    ll = len(values)
    mid = int(ll/2)
    values = sorted(values)
    if ll % 2 == 0:
        return (values[mid - 1] + values[mid])/2
    else:
        return values[mid]


def update_counter(curr, new):
    if new is None:
        return curr
    if curr is None:
        curr = collections.Counter()
    if isinstance(new, str):
        new = [new]
    if not isinstance(curr, collections.Counter):
        curr = collections.Counter(curr)
    curr.update(new)
    return curr


# Aggregators
Aggregator = collections.namedtuple('Aggregator',
                                    ['func', 'finaliser', 'dataType', 'copyProperties'])
AGGREGATORS = {
    'sum': Aggregator(lambda curr, new:
                      new + curr if curr is not None else new,
                      identity,
                      None,
                      False),
    'avg': Aggregator(lambda curr, new:
                      (curr[0] + 1, new + curr[1])
                      if curr is not None
                      else (1, new),
                      lambda value: value[1] / value[0],
                      None,
                      False),
    'median': Aggregator(lambda curr, new:
                         curr + [new] if curr is not None else [new],
                         median,
                         None,
                         True),
    'max': Aggregator(lambda curr, new:
                      max(new, curr) if curr is not None else new,
                      identity,
                      None,
                      False),
    'min': Aggregator(lambda curr, new:
                      min(new, curr) if curr is not None else new,
                      identity,
                      None,
                      False),
    'first': Aggregator(lambda curr, new:
                        curr if curr is not None else new,
                        identity,
                        None,
                        True),
    'last': Aggregator(lambda curr, new: new,
                       identity,
                       None,
                       True),
    'count': Aggregator(lambda curr, new:
                        curr+1 if curr is not None else 1,
                        identity,
                        'integer',
                        False),
    'any': Aggregator(lambda curr, new: new,
                      identity,
                      None,
                      True),
    'set': Aggregator(lambda curr, new:
                      curr.union({new}) if curr is not None else {new},
                      lambda value: list(value) if value is not None else [],
                      'array',
                      False),
    'array': Aggregator(lambda curr, new:
                        curr + [new] if curr is not None else [new],
                        lambda value: value if value is not None else [],
                        'array',
                        False),
    'counters': Aggregator(lambda curr, new:
                           update_counter(curr, new),
                           lambda value:
                           list(collections.Counter(value).most_common()) if value is not None else [],
                           'array',
                           False),
}


# Input helpers

def fix_fields(fields):
    for field in sorted(fields.keys()):
        spec = fields[field]
        if spec is None:
            fields[field] = spec = {}
        if 'name' not in spec:
            spec['name'] = field
        if 'aggregate' not in spec:
            spec['aggregate'] = 'any'
    return fields


def expand_fields(fields, schema_fields):
    if '*' in fields:
        existing_names = set(f['name'] for f in fields.values())
        spec = fields.pop('*')
        for sf in schema_fields:
            sf_name = sf['name']
            if sf_name not in existing_names:
                fields[sf_name] = copy.deepcopy(spec)
                fields[sf_name]['name'] = sf_name


def order_fields(fields, schema_fields):
    ordered_fields = collections.OrderedDict()
    for descriptor in schema_fields:
        name = descriptor['name']
        if name in fields:
            ordered_fields[name] = fields.pop(name)
    for name in sorted(fields.keys()):
        ordered_fields[name] = fields[name]
    return ordered_fields


def concatenator(resources, all_target_fields, field_mapping):
    for resource_ in resources:
        for row in resource_:
            processed = dict((k, '') for k in all_target_fields)
            values = [(field_mapping[k], v) for (k, v)
                    in row.items()
                    if k in field_mapping]
            assert len(values) > 0
            processed.update(dict(values))
            yield processed


def join_aux(source_name, source_key, source_delete,  # noqa: C901
             target_name, target_key, fields, full, mode):

    deduplication = target_key is None
    fields = fix_fields(fields)
    source_key = KeyCalc(source_key)
    target_key = KeyCalc(target_key) if target_key is not None else target_key
    # We will store db keys as boolean flags:
    # - False -> inserted/not used
    # - True -> inserted/used
    db_keys_usage = KVFile()
    db = KVFile()

    # Mode of join operation
    if full is not None:
        warnings.warn(
            'For the `join` processor the `full=True` flag is deprecated. '
            'Please use the "mode" parameter instead.',
            UserWarning)
        mode = 'half-outer' if full else 'inner'
    assert mode in ['inner', 'half-outer', 'full-outer']

    # Indexes the source data
    def indexer(resource):
        for row_number, row in enumerate(resource, start=1):
            key = source_key(row, row_number)
            try:
                current = db.get(key)
            except KeyError:
                current = {}
            for field, spec in fields.items():
                name = spec['name']
                curr = current.get(field)
                agg = spec['aggregate']
                if agg != 'count':
                    new = row.get(name)
                else:
                    new = ''
                if new is not None:
                    current[field] = AGGREGATORS[agg].func(curr, new)
                elif field not in current:
                    current[field] = None
            if mode == 'full-outer':
                current['__key__'] = [row.get(field) for field in source_key.key_list]
            db.set(key, current)
            db_keys_usage.set(key, False)
            yield row

    # Generates the joined data
    def process_target(resource):
        if deduplication:
            # just empty the iterable
            collections.deque(indexer(resource), maxlen=0)
            for key, value in db.items():
                row = dict(
                    (f, None) for f in fields.keys()
                )
                row.update(dict(
                    (k, AGGREGATORS[fields[k]['aggregate']].finaliser(v))
                    for k, v in value.items()
                ))
                yield row
        else:
            for row_number, row in enumerate(resource, start=1):
                key = target_key(row, row_number)
                try:
                    extra = create_extra_by_key(key)
                    db_keys_usage.set(key, True)
                except KeyError:
                    if mode == 'inner':
                        continue
                    extra = dict(
                        (k, row.get(k))
                        for k in fields.keys()
                    )
                row.update(extra)
                yield row
            if mode == 'full-outer':
                for key, value in db_keys_usage.items():
                    if value is False:
                        extra = create_extra_by_key(key)
                        yield extra

    # Creates extra by key
    def create_extra_by_key(key):
        extra = db.get(key)
        key = extra.pop('__key__', None)
        extra = dict(
            (k, AGGREGATORS[fields[k]['aggregate']].finaliser(v))
            for k, v in extra.items()
            if k in fields
        )
        if key:
            for k, v in zip(target_key.key_list, key):
                extra[k] = v
        return extra

    # Yields the new resources
    def new_resource_iterator(resource_iterator):
        has_index = False
        for resource in resource_iterator:
            name = resource.res.name
            if name == source_name:
                has_index = True
                if source_delete:
                    # just empty the iterable
                    collections.deque(indexer(resource), maxlen=0)
                else:
                    yield indexer(resource)
                if deduplication:
                    yield process_target(resource)
            elif name == target_name:
                assert has_index
                yield process_target(resource)
            else:
                yield resource

    # Updates / creates the target resource descriptor
    def process_target_resource(source_spec, resource):
        target_fields = \
            resource.setdefault('schema', {}).setdefault('fields', [])
        for name, spec in fields.items():
            agg = spec['aggregate']
            data_type = AGGREGATORS[agg].dataType
            copy_properties = AGGREGATORS[agg].copyProperties
            to_copy = {}
            if data_type is None:
                try:
                    source_field = \
                        next(filter(lambda f: f['name'] == spec['name'],
                                    source_spec['schema']['fields']))
                except StopIteration:
                    raise KeyError('Failed to find field with name %s in resource %s' %
                                   (spec['name'], source_spec['name']))
                if copy_properties:
                    to_copy = copy.deepcopy(source_field)
                data_type = source_field['type']
            try:
                existing_field = next(iter(filter(
                    lambda f: f['name'] == name,
                    target_fields)))
                assert existing_field['type'] == data_type, \
                    'Reusing %s but with different data types: %s != %s' % (name, existing_field['type'], data_type)
            except StopIteration:
                to_copy.update({
                    'name': name,
                    'type': data_type
                })
                target_fields.append(to_copy)
        return resource

    # Updates the datapackage descriptor based on parameters
    def process_datapackage(datapackage):

        new_resources = []
        source_spec = None

        resource_names = [resource['name'] for resource in datapackage['resources']]
        assert source_name in resource_names, \
            'Source resource ({}) not found package (target={}, found: {})'\
            .format(source_name, target_name, resource_names)
        assert target_name in resource_names, \
            'Target resource ({}) not found package (source={}, found: {})'\
            .format(target_name, source_name, resource_names)

        for resource in datapackage['resources']:

            if resource['name'] == source_name:
                nonlocal fields
                source_spec = resource
                schema_fields = source_spec.get('schema', {}).get('fields', [])
                expand_fields(fields, schema_fields)
                fields = order_fields(fields, schema_fields)
                if not source_delete:
                    new_resources.append(resource)
                if deduplication:
                    resource = process_target_resource(
                        source_spec,
                        {
                            'name': target_name,
                            'path': os.path.join('data', target_name + '.csv')
                        })
                    new_resources.append(resource)

            elif resource['name'] == target_name:
                assert isinstance(source_spec, dict),\
                       'Source resource ({}) must appear before target resource ({}), found: {}'\
                       .format(source_name, target_name, resource_names)
                resource = process_target_resource(source_spec, resource)
                new_resources.append(resource)

            else:
                new_resources.append(resource)

        datapackage['resources'] = new_resources

    def func(package: PackageWrapper):
        process_datapackage(package.pkg.descriptor)
        yield package.pkg
        yield from new_resource_iterator(package)

    return func


def join(source_name, source_key, target_name, target_key, fields={}, full=None, mode='half-outer', source_delete=True):
    return join_aux(source_name, source_key, source_delete, target_name, target_key, fields, full, mode)


def join_with_self(resource_name, join_key, fields):
    return join_aux(resource_name, join_key, True, resource_name, None, fields, True, None)


def join_self(source_name, source_key, target_name, fields):
    import warnings
    warnings.warn('join_self is being deprecated, use join_with_self instead',
                  DeprecationWarning)
    return join_aux(source_name, source_key, True, target_name, None, fields, True, None)



import collections
from dataflows import PackageWrapper
from dataflows.helpers.resource_matcher import ResourceMatcher


def delete_resource(resources):

    def func(package: PackageWrapper):
        matcher = ResourceMatcher(resources, package.pkg)
        descriptor = package.pkg.descriptor
        descriptor['resources'] = [
            resource for resource in descriptor['resources']
            if not matcher.match(resource['name'])
        ]
        package.pkg.commit()
        yield package.pkg

        for r in package:
            if not matcher.match(r.res.name):
                yield r
            else:
                collections.deque(r, maxlen=0)

    return func



import datetime
import json

import decimal
import isodate

# On some platforms, pre year 1000 dates are not formatted using 4 digits (which is later unparseable)
try:
    if datetime.date(1, 1, 1).strftime('%04Y') == '4Y':
        DATE_F_FORMAT = '%Y-%m-%d'
        DATETIME_F_FORMAT = '%Y-%m-%dT%H:%M:%S'
    else:
        DATE_F_FORMAT = '%04Y-%m-%d'
        DATETIME_F_FORMAT = '%04Y-%m-%dT%H:%M:%S'
except ValueError:
    DATE_F_FORMAT = '%Y-%m-%d'
    DATETIME_F_FORMAT = '%Y-%m-%dT%H:%M:%S'

DATE_FORMAT = DATE_F_FORMAT
DATETIME_FORMAT = DATETIME_F_FORMAT
TIME_FORMAT = '%H:%M:%S'

DATE_P_FORMAT = '%Y-%m-%d'
DATETIME_P_FORMAT = '%Y-%m-%dT%H:%M:%S'
TIME_P_FORMAT = TIME_F_FORMAT = TIME_FORMAT


class CommonJSONDecoder(json.JSONDecoder):
    """
    Common JSON Encoder
    json.loads(myString, cls=CommonJSONEncoder)
    """

    @classmethod
    def object_hook(cls, obj):
        if 'type{decimal}' in obj:
            try:
                return decimal.Decimal(obj['type{decimal}'])
            except decimal.InvalidOperation:
                pass
        if 'type{time}' in obj:
            try:
                return datetime.datetime \
                    .strptime(obj['type{time}'], TIME_P_FORMAT) \
                    .time()
            except ValueError:
                pass
        if 'type{datetime}' in obj:
            try:
                (isoformat, tzofs, tzname) = obj['type{datetime}']
                parsed = datetime.datetime \
                    .strptime(isoformat, DATETIME_P_FORMAT)
                if tzname is not None:
                    return datetime.datetime \
                        .combine(parsed.date(), parsed.time(),
                                 datetime.timezone(datetime.timedelta(seconds=tzofs), tzname))
                else:
                    return parsed
            except ValueError:
                pass
        if 'type{date}' in obj:
            try:
                return datetime.datetime \
                    .strptime(obj['type{date}'], DATE_P_FORMAT) \
                    .date()
            except ValueError:
                pass
        if 'type{duration}' in obj:
            try:
                return isodate.parse_duration(obj['type{duration}'])
            except ValueError:
                pass
        if 'type{set}' in obj:
            try:
                return set(obj['type{set}'])
            except ValueError:
                pass

        return obj

    def __init__(self, **kwargs):
        kwargs['object_hook'] = self.object_hook
        super(CommonJSONDecoder, self).__init__(**kwargs)


class CommonJSONEncoder(json.JSONEncoder):
    """
    Common JSON Encoder
    json.dumps(myString, cls=CommonJSONEncoder)
    """

    def default(self, obj):

        if isinstance(obj, decimal.Decimal):
            return {'type{decimal}': str(obj)}
        elif isinstance(obj, datetime.time):
            return {'type{time}': obj.strftime(TIME_F_FORMAT)}
        elif isinstance(obj, datetime.datetime):
            return {'type{datetime}':
                    (obj.strftime(DATETIME_F_FORMAT),
                     obj.utcoffset().seconds if obj.utcoffset() is not None else None,
                     obj.tzname())}
        elif isinstance(obj, datetime.date):
            return {'type{date}': obj.strftime(DATE_F_FORMAT)}
        elif isinstance(obj, (isodate.Duration, datetime.timedelta)):
            return {'type{duration}': isodate.duration_isoformat(obj)}
        elif isinstance(obj, set):
            return {'type{set}': list(obj)}
        return super().default(obj)


class ejson():

    @staticmethod
    def dumps(*args, **kwargs):
        kwargs['cls'] = CommonJSONEncoder
        return json.dumps(*args, **kwargs)

    @staticmethod
    def loads(*args, **kwargs):
        kwargs['cls'] = CommonJSONDecoder
        return json.loads(*args, **kwargs)

    @staticmethod
    def dump(*args, **kwargs):
        kwargs['cls'] = CommonJSONEncoder
        return json.dump(*args, **kwargs)

    @staticmethod
    def load(*args, **kwargs):
        kwargs['cls'] = CommonJSONDecoder
        return json.load(*args, **kwargs)



import zipfile

from .file_dumper import FileDumper


class ZipDumper(FileDumper):

    def __init__(self, out_file, **options):
        super(ZipDumper, self).__init__(options)
        self.out_file = open(out_file, 'wb')
        self.zip_file = zipfile.ZipFile(self.out_file, 'w')

    def write_file_to_output(self, filename, path):
        self.zip_file.write(filename, arcname=path,
                            compress_type=zipfile.ZIP_DEFLATED)

    def finalize(self):
        self.zip_file.close()
        if not self.out_file.closed:
            self.out_file.close()
        super(ZipDumper, self).finalize()



from .row_processor import row_processor
from .rows_processor import rows_processor
from .datapackage_processor import datapackage_processor
from .iterable_loader import iterable_loader
from .resource_matcher import ResourceMatcher



from .. import DataStreamProcessor


class resources_processor(DataStreamProcessor):

    def __init__(self, rows_processor_func):
        super(resources_processor, self).__init__()
        self.func = rows_processor_func

    def process_resources(self, resources):
        yield from self.func(resources)



from inspect import signature

import re

from ..helpers.resource_matcher import ResourceMatcher
from .. import DataStreamProcessor, schema_validator


class set_type(DataStreamProcessor):

    def __init__(self, name, resources=-1, regex=True, on_error=None, transform=None, **options):
        super(set_type, self).__init__()
        if not regex:
            name = re.escape(name)
        self.name = re.compile(f'^{name}$')
        self.options = options
        self.resources = resources
        self.field_names = dict()
        self.on_error = on_error
        self.transform = self.wrap_transformer(transform) if transform else None

    def wrap_transformer(self, transform):
        assert callable(transform)
        try:
            sig = signature(transform).parameters
        except Exception:
            sig = set()

        def func(v, field_name=None, row=None):
            kw = {}
            if 'row' in sig:
                kw['row'] = row
            if 'field_name' in sig:
                kw['field_name'] = field_name
            return transform(v, **kw)
        return func

    def transformer(self, rows, field_names):
        for row in rows:
            for field_name in field_names:
                row[field_name] = self.transform(row.get(field_name), field_name=field_name, row=row)
            yield row

    def process_resources(self, resources):
        for res in resources:
            if self.matcher.match(res.res.name):
                field_names = self.field_names.get(res.res.name, [])
                if len(field_names) > 0:
                    it = res
                    if self.transform is not None:
                        it = self.transformer(it, field_names)
                    yield schema_validator(res.res, it,
                                           field_names=field_names,
                                           on_error=self.on_error)
                else:
                    yield res
            else:
                yield res

    def process_datapackage(self, dp):
        dp = super(set_type, self).process_datapackage(dp)
        self.matcher = ResourceMatcher(self.resources, dp)
        added = False
        for res in dp.descriptor['resources']:
            if self.matcher.match(res['name']):
                for field in res['schema']['fields']:
                    if self.name.match(field['name']):
                        field.update(self.options)
                        self.field_names.setdefault(res['name'], []).append(field['name'])
                        added = True
        assert added, 'Failed to find field {} in schema'.format(self.name)
        return dp



from .. import DataStreamProcessor, PackageWrapper


class datapackage_processor(DataStreamProcessor):

    def __init__(self, dp_processor_func):
        super(datapackage_processor, self).__init__()
        self.func = dp_processor_func
        self.dp = None
        self.dp_processor = None

    def process_datapackage(self, dp):
        self.dp = PackageWrapper(dp)
        self.dp_processor = self.func(self.dp)
        ret = next(self.dp_processor)
        if ret is None:
            return dp
        return ret

    def process_resources(self, res_iter):
        self.dp.it = res_iter
        yield from self.dp_processor



import os
import json
import tempfile
import hashlib
from typing import Dict
from dataflows.base.resource_wrapper import ResourceWrapper

from datapackage import Resource

from .dumper_base import DumperBase
from .formats import CSVFormat, JSONFormat, GeoJSONFormat, ExcelFormat, FileFormat


# see https://stackoverflow.com/questions/7150826/how-can-i-get-the-default-file-permissions-in-python
def UmaskNamedTemporaryFile(*args, **kargs):
    fdesc = tempfile.NamedTemporaryFile(*args, **kargs)
    # we need to set umask to get its current value. As noted
    # by Florian Brucker (comment), this is a potential security
    # issue, as it affects all the threads. Considering that it is
    # less a problem to create a file with permissions 000 than 666,
    # we use 666 as the umask temporary value.
    umask = os.umask(0o666)
    os.umask(umask)
    os.chmod(fdesc.name, 0o666 & ~umask)
    return fdesc


class FileDumper(DumperBase):

    def __init__(self, options: dict):
        super(FileDumper, self).__init__(options)
        self.force_format = options.pop('force_format', True)
        self.forced_format = options.pop('format', 'csv')
        self.temporal_format_property = options.pop('temporal_format_property', None)
        self.use_titles = options.pop('use_titles', False)
        self.writer_options = options.pop('options', dict())

    def process_datapackage(self, datapackage):
        datapackage = \
            super(FileDumper, self).process_datapackage(datapackage)

        self.file_formatters: Dict[str, FileFormat] = {}

        # Make sure all resources are proper CSVs
        resource: Resource = None
        for i, resource in enumerate(datapackage.resources):
            if self.force_format:
                file_format = self.forced_format
            else:
                _, file_format = os.path.splitext(resource.source)
                file_format = file_format[1:]
            file_formatter = {
                'csv': CSVFormat,
                'json': JSONFormat,
                'geojson': GeoJSONFormat,
                'excel': ExcelFormat,
                'xlsx': ExcelFormat,
            }.get(file_format)
            if file_format is not None:
                self.file_formatters[resource.name] = file_formatter
                self.file_formatters[resource.name].prepare_resource(resource)
                resource.commit()
                datapackage.descriptor['resources'][i] = resource.descriptor

        return datapackage

    def handle_datapackage(self):

        # Handle temporal_format_property
        if self.temporal_format_property:
            for resource in self.datapackage.descriptor['resources']:
                for field in resource['schema']['fields']:
                    if field.get('type') in ['datetime', 'date', 'time']:
                        format = field.pop(self.temporal_format_property, None)
                        if format:
                            field['format'] = format
            self.datapackage.commit()

        temp_file = UmaskNamedTemporaryFile(mode='w+', delete=False, encoding='utf-8')
        indent = 2 if self.pretty_descriptor else None
        json.dump(self.datapackage.descriptor, temp_file, indent=indent, sort_keys=True, ensure_ascii=False)
        temp_file_name = temp_file.name
        filesize = temp_file.tell()
        temp_file.close()
        DumperBase.inc_attr(self.datapackage.descriptor, self.datapackage_bytes, filesize)
        self.write_file_to_output(temp_file_name, 'datapackage.json')
        # if location is not None:
        #     stats.setdefault(STATS_DPP_KEY, {})[STATS_OUT_DP_URL_KEY] = location
        os.unlink(temp_file_name)
        super(FileDumper, self).handle_datapackage()

    def write_file_to_output(self, filename, path):
        raise NotImplementedError()

    def rows_processor(self, resource, writer, temp_file):
        for row in resource:
            writer.write_row(row)
            yield row
        writer.finalize_file()

        # Get resource descriptor
        resource_descriptor = resource.res.descriptor
        for descriptor in self.datapackage.descriptor['resources']:
            if descriptor['name'] == resource.res.descriptor['name']:
                resource_descriptor = descriptor

        # File size:
        filesize = temp_file.tell()
        DumperBase.inc_attr(self.datapackage.descriptor, self.datapackage_bytes, filesize)
        DumperBase.inc_attr(resource_descriptor, self.resource_bytes, filesize)

        # File Hash:
        if self.resource_hash:
            hasher = FileDumper.hash_handler(temp_file)
            # Update path with hash
            if self.add_filehash_to_path:
                DumperBase.insert_hash_in_path(resource_descriptor, hasher.hexdigest())
            DumperBase.set_attr(resource_descriptor, self.resource_hash, hasher.hexdigest())

        # Finalise
        filename = temp_file.name
        temp_file.close()
        self.write_file_to_output(filename, resource.res.source)
        os.unlink(filename)

    def process_resource(self, resource: ResourceWrapper):
        if resource.res.name in self.file_formatters:
            schema = resource.res.schema

            file_formatter = self.file_formatters[resource.res.name]

            temp_file = UmaskNamedTemporaryFile(
                mode=file_formatter.FILE_MODE, delete=False,
                newline='' if 'b' not in file_formatter.FILE_MODE else None
            )
            writer_kwargs = self.writer_options
            if self.use_titles:
                writer_kwargs['use_titles'] = True
            writer_kwargs['temporal_format_property'] = self.temporal_format_property
            writer_kwargs['resource'] = resource.res
            writer = file_formatter(temp_file, schema, **writer_kwargs)

            return self.rows_processor(resource,
                                       writer,
                                       temp_file)
        else:
            return resource

    @staticmethod
    def hash_handler(tfile):
        tfile.seek(0)
        hasher = hashlib.md5()
        data = 'x'
        while len(data) > 0:
            data = tfile.read(1024)
            if isinstance(data, str):
                hasher.update(data.encode('utf8'))
            elif isinstance(data, bytes):
                hasher.update(data)
        return hasher



import itertools
import os
import multiprocessing as mp
import threading
import queue

from ..helpers import ResourceMatcher
from .. import PackageWrapper, ResourceWrapper


def init_mp(num_processors, row_func, q_in, q_internal):
    q_out = mp.Queue()
    processes = [mp.Process(target=work, args=(q_in, q_out, row_func)) for _ in range(num_processors)]
    for process in processes:
        process.start()
    t_fetch = threading.Thread(target=fetcher, args=(q_out, q_internal, num_processors))
    t_fetch.start()
    return (processes, t_fetch)


def fini_mp(processes, t_fetch):
    for process in processes:
        try:
            process.join(timeout=10)
        except Exception:
            try:
                process.kill()
            except Exception:
                pass
        finally:
            if hasattr(process, 'close'):
                process.close()
    t_fetch.join()


def producer(res, q_in, q_internal, num_processors, predicate):
    try:
        for row in res:
            if predicate(row):
                q_in.put(row)
            else:
                q_internal.put(row)
        for _ in range(num_processors):
            q_in.put(None)
    except Exception:
        q_internal.put(None)
        return 1
    return 0


def fetcher(q_out, q_internal, num_processors):
    expected_nones = num_processors
    while True:
        row = q_out.get()
        if row is None:
            expected_nones -= 1
            if expected_nones == 0:
                q_internal.put(None)
                break
            continue
        q_internal.put(row)


def work(q_in: mp.Queue, q_out: mp.Queue, row_func):
    pid = os.getpid()
    try:
        while True:
            row = q_in.get()
            if row is None:
                break
            try:
                row_func(row)
            except Exception as e:
                print(pid, 'FAILED TO RUN row_func {}\n'.format(e))
                pass
            q_out.put(row)
    except Exception:
        pass
    finally:
        q_out.put(None)


def fork(res, row_func, num_processors, predicate):
    predicate = predicate or (lambda x: True)
    for row in res:
        if predicate(row):
            res = itertools.chain([row], res)
            q_in = mp.Queue()
            q_internal = queue.Queue()
            t_prod = threading.Thread(target=producer, args=(res, q_in, q_internal, num_processors, predicate))
            t_prod.start()

            processes, t_fetch = init_mp(num_processors, row_func, q_in, q_internal)

            while True:
                row = q_internal.get()
                if row is None:
                    break
                yield row
            t_prod.join()
            fini_mp(processes, t_fetch)
        else:
            yield row


def parallelize(row_func, num_processors=None, resources=None, predicate=None):
    num_processors = num_processors or 2*os.cpu_count()

    def func(package: PackageWrapper):
        yield package.pkg
        matcher = ResourceMatcher(resources, package.pkg)

        res: ResourceWrapper
        for res in package:
            if matcher.match(res.res.name):
                yield fork(res, row_func, num_processors, predicate)
            else:
                yield res

    return func



import functools
import collections

from ..helpers.resource_matcher import ResourceMatcher

Aggregator = collections.namedtuple('Aggregator', ['func'])

AGGREGATORS = {
    'sum': Aggregator(lambda values, fstr, row: sum(values)),
    'avg': Aggregator(lambda values, fstr, row: sum(values) / len(values)),
    'max': Aggregator(lambda values, fstr, row: max(values)),
    'min': Aggregator(lambda values, fstr, row: min(values)),
    'multiply': Aggregator(
        lambda values, fstr, row: functools.reduce(lambda x, y: x*y, values)),
    'constant': Aggregator(lambda values, fstr, row: fstr),
    'join': Aggregator(
        lambda values, fstr, row: fstr.join([str(x) for x in values])),
    'format': Aggregator(lambda values, fstr, row: fstr.format(**row)),
}


def get_type(res_fields, operation_fields, operation):
    types = [f.get('type') for f in res_fields if f['name'] in operation_fields]
    if 'any' in types:
        return 'any'
    if operation in ('format', 'join'):
        return 'string'
    if ('number' in types) or (operation == 'avg'):
        return 'number'
    # integers
    if len(types):
        return types[0]
    # constant
    return 'any'


def process_resource(fields, rows):
    for row in rows:
        for field in fields:
            op = field['operation']
            target = field['target']['name']
            if isinstance(op, str):
                values = [
                    row.get(c)
                    for c in field.get('source', [])
                    if row.get(c) is not None
                ]
                with_ = field.get('with', field.get('with_', ''))
                new_col = AGGREGATORS[op].func(values, with_, row)
                row[target] = new_col
            elif callable(op):
                row[target] = op(row)
        yield row


def get_new_fields(resource, fields):
    new_fields = []
    for f in fields:
        target = f['target']
        if isinstance(target, str):
            target = dict(
                name=target,
                type=get_type(resource['schema']['fields'],
                              f.get('source', []),
                              f['operation'])
            )
            new_fields.append(target)
        elif isinstance:
            new_fields.append(target)
    return new_fields


def add_computed_field(*args, resources=None, **kw):

    def func(package):
        matcher = ResourceMatcher(resources, package.pkg)
        assert len(args) < 2, 'add_computed_fields expects at most one positional argument'
        if len(args) == 0:
            fields = [kw]
        elif len(args) == 1:
            fields = args[0]

        for resource in package.pkg.descriptor['resources']:
            if matcher.match(resource['name']):
                new_fields = get_new_fields(resource, fields)
                resource['schema']['fields'].extend(new_fields)
        yield package.pkg

        for f in fields:
            target = f['target']
            if isinstance(target, str):
                f['target'] = dict(name=target)

        for resource in package:
            if not matcher.match(resource.res.name):
                yield resource
            else:
                yield process_resource(fields, resource)

    return func



from tabulate import tabulate
from ..helpers.resource_matcher import ResourceMatcher


try:
    from IPython.core.display import display, HTML
    get_ipython
    def display_html(data):
        display(HTML(data))
except (NameError, ImportError):
    def display_html(data):
        print(data)


def _header_print(header, kwargs):
    if kwargs.get('tablefmt') == 'html':
        display_html(f'<h3>{header}</h3>')
    else:
        print(f'{header}:')


def _table_print(data, kwargs):
    if kwargs.get('tablefmt') == 'html':
        display_html(data)
    else:
        print(data)


def truncate_cell(value, max_size):
    value = str(value)
    if max_size is not None and len(value) > max_size:
        return value[:max_size] + ' ...'
    else:
        return value


def printer(num_rows=10, last_rows=None, fields=None, resources=None,
            header_print=_header_print, table_print=_table_print, max_cell_size=100, **kwargs):

    def func(rows):
        spec = rows.res

        if not ResourceMatcher(resources, spec.descriptor).match(spec.name):
            yield from rows
            return

        header_print(spec.name, kwargs)

        schema_fields = spec.schema.fields
        if fields:
            schema_fields = [f for f in schema_fields if f.name in fields]

        field_names = [f.name for f in schema_fields]
        headers = ['#'] + [
            '{}\n({})'.format(f.name, f.type) for f in schema_fields
        ]
        toprint = []
        last = []
        x = 1

        for i, row in enumerate(rows):

            index = i + 1
            prow = [index] + [truncate_cell(row.get(f), max_cell_size) for f in field_names]
            yield row

            if index - x == (num_rows + 1):
                x *= num_rows

            if 0 <= index - x <= num_rows:
                last.clear()
                if toprint and toprint[-1][0] != index - 1:
                    toprint.append(['...'])
                toprint.append(prow)
            else:
                last.append(prow)
                if len(last) > (last_rows or num_rows):
                    last = last[1:]

        if toprint and last and toprint[-1][0] != last[0][0] - 1:
            toprint.append(['...'])

        toprint += last

        table_print(tabulate(toprint, headers=headers, **kwargs), kwargs)

    return func



from .add_computed_field import add_computed_field


def add_field(name, type, default=None, resources=None, **options):
    return add_computed_field(
        target=dict(
            name=name,
            type=type,
            **options
        ),
        resources=resources,
        operation=(
            default
            if callable(default) else
            (lambda row: default)
        )
    )



import itertools

from ..helpers.resource_matcher import ResourceMatcher


def concatenator(resources, all_target_fields, field_mapping):
    for resource_ in resources:
        for row in resource_:
            processed = dict((k, None) for k in all_target_fields)
            values = [(field_mapping[k], v) for (k, v)
                      in row.items()
                      if k in field_mapping and v is not None]
            if len(values) == 0:
                message = 'Got an empty row after concatenation' +\
                    '(resource=%s, source=%r)' % (resource_.res.name, row)
                assert len(values) > 0, message

            processed.update(dict(values))
            yield processed


def concatenate(fields, target={}, resources=None):

    def func(package):
        matcher = ResourceMatcher(resources, package.pkg)
        # Prepare target resource
        if 'name' not in target:
            target['name'] = 'concat'
        if 'path' not in target:
            target['path'] = 'data/' + target['name'] + '.csv'
        target.update(dict(
            mediatype='text/csv',
            schema=dict(fields=[], primaryKey=[]),
            profile='tabular-data-resource'
        ))

        # Create mapping between source field names to target field names
        field_mapping = {}
        for target_field, source_fields in fields.items():
            if source_fields is not None:
                for source_field in source_fields:
                    if source_field in field_mapping:
                        raise RuntimeError('Duplicate appearance of %s (%r)' % (source_field, field_mapping))
                    field_mapping[source_field] = target_field

            if target_field in field_mapping:
                raise RuntimeError('Duplicate appearance of %s' % target_field)

            field_mapping[target_field] = target_field

        # Create the schema for the target resource
        needed_fields = list(fields.keys())
        for resource in package.pkg.descriptor['resources']:
            if not matcher.match(resource['name']):
                continue

            schema = resource.get('schema', {})
            pk = schema.get('primaryKey', [])
            for field in schema.get('fields', []):
                orig_name = field['name']
                if orig_name in field_mapping:
                    name = field_mapping[orig_name]
                    if name not in needed_fields:
                        continue
                    if orig_name in pk:
                        target['schema']['primaryKey'].append(name)
                    target['schema']['fields'].append(field)
                    field['name'] = name
                    needed_fields.remove(name)

        if len(target['schema']['primaryKey']) == 0:
            del target['schema']['primaryKey']

        for name in needed_fields:
            target['schema']['fields'].append(dict(
                name=name, type='string'
            ))

        # Update resources in datapackage (make sure they are consecutive)
        prefix = True
        suffix = False
        num_concatenated = 0
        new_resources = []
        for resource in package.pkg.descriptor['resources']:
            name = resource['name']
            match = matcher.match(name)
            if prefix:
                if match:
                    prefix = False
                    num_concatenated += 1
                else:
                    new_resources.append(resource)
            elif suffix:
                assert not match
                new_resources.append(resource)
            else:
                if not match:
                    suffix = True
                    new_resources.append(target)
                    new_resources.append(resource)
                else:
                    num_concatenated += 1
        if not suffix:
            new_resources.append(target)

        package.pkg.descriptor['resources'] = new_resources
        yield package.pkg

        needed_fields = list(fields.keys())
        it = iter(package)
        for resource in it:
            if matcher.match(resource.res.name):
                resource_chain = \
                    itertools.chain([resource],
                                    itertools.islice(it,
                                                     num_concatenated-1))
                yield concatenator(resource_chain, needed_fields, field_mapping)
            else:
                yield resource

    return func



import json
from datapackage.resource import Resource
import isodate
import openpyxl
from pathlib import Path

from dataflows.helpers.extended_json import (
    DATETIME_P_FORMAT, DATE_P_FORMAT, TIME_P_FORMAT,
)

from .base import FileFormat, identity, comma_separated, json_dumps


class ExcelFormat(FileFormat):

    SERIALIZERS = {
        'array': comma_separated,
        'object': json_dumps,
        'datetime': identity,
        'date': identity,
        'time': identity,
        'duration': lambda d: isodate.duration_isoformat(d),
        'geopoint': lambda d: '{}, {}'.format(*d),
        'geojson': json.dumps,
        'year': identity,
        'yearmonth': lambda d: '{:04d}-{:02d}'.format(*d),
    }
    NULL_VALUE = ''

    PYTHON_DIALECT = {
        'number': {
            'decimalChar': '.',
            'groupChar': ''
        },
        'date': {
            'format': DATE_P_FORMAT
        },
        'time': {
            'format': TIME_P_FORMAT
        },
        'datetime': {
            'format': DATETIME_P_FORMAT
        },
        'boolean': {
            'trueValues': ['True'],
            'falseValues': ['False']
        }
    }
    FILE_MODE = 'wb+'

    def __init__(self, file, schema, use_titles=False, **options):
        self.resource: Resource = options['resource']
        self.tmpfile = file
        self.sheetname = options.get('sheetname') or self.resource.name or 'Sheet1'
        self.update_existing = options.get('update_existing')
        try:
            if self.update_existing and Path(self.update_existing).exists():
                self.workbook = openpyxl.load_workbook(self.update_existing)
            else:
                self.workbook = openpyxl.Workbook()
                for sheet in self.workbook.worksheets:
                    self.workbook.remove(sheet)
            if self.sheetname in self.workbook.sheetnames:
                self.workbook.remove(self.workbook[self.sheetname])
            worksheet = self.workbook.create_sheet(self.sheetname)
            self.workbook.save(filename=self.tmpfile.name)
        finally:
            self.workbook.close()

        super(ExcelFormat, self).__init__(worksheet, schema, **options)

        if use_titles:
            headers = dict((f.name, f.descriptor.get('title', f.name)) for f in schema.fields)
        else:
            headers = dict((f.name, f.name) for f in schema.fields)
        self.write_transformed_row(headers)

    @classmethod
    def prepare_resource(cls, resource):
        descriptor = resource.descriptor
        descriptor['path'] = str(Path(descriptor['path']).with_suffix('.xlsx'))
        descriptor['format'] = 'xlsx'
        super(ExcelFormat, cls).prepare_resource(resource)

    def write_transformed_row(self, transformed_row):
        self.writer.append(transformed_row[k] for k in self.fields)

    def finalize_file(self):
        self.workbook.save(self.tmpfile.name)
        self.workbook.close()



from .. import DataStreamProcessor


class rows_processor(DataStreamProcessor):

    def __init__(self, rows_processor_func):
        super(rows_processor, self).__init__()
        self.func = rows_processor_func

    def process_resource(self, resource):
        yield from self.func(resource)



from datapackage import Resource


class ResourceWrapper:
    def __init__(self, res: Resource, it):
        self.res = res
        self.it = it
        assert res is not None
        assert isinstance(res, Resource)

    def __iter__(self):
        return self.it



from tabulator.parser import Parser
from tabulator.helpers import reset_stream


class XMLParser(Parser):
    options = []

    def __init__(self, loader, force_parse, **options):
        self.__loader = loader
        self.__force_parse = force_parse
        self.__extended_rows = None
        self.__encoding = None
        self.__chars = None

    def open(self, source, encoding=None):
        self.close()
        self.__chars = self.__loader.load(source, encoding=encoding)
        self.__encoding = getattr(self.__chars, 'encoding', encoding)
        if self.__encoding:
            self.__encoding.lower()
        self.reset()

    def close(self):
        if not self.closed:
            self.__chars.close()

    def reset(self):
        reset_stream(self.__chars)
        self.__extended_rows = self.__iter_extended_rows()

    @property
    def closed(self):
        return self.__chars is None or self.__chars.closed

    @property
    def encoding(self):
        return self.__encoding

    @property
    def extended_rows(self):
        return self.__extended_rows

    # Private

    def __iter_extended_rows(self):
        from xml.etree.ElementTree import parse
        from xmljson import parker

        parsed = parker.data(parse(self.__chars).getroot())
        elements = list(parsed.values())
        if len(elements) > 0:
            elements = elements[0]
        else:
            elements = []
        for row_number, row in enumerate(elements, start=1):
            keys, values = zip(*(row.items()))
            yield (row_number, list(keys), list(values))



import json
import isodate
from pathlib import Path

from dataflows.helpers.extended_json import (
    DATETIME_F_FORMAT, DATE_F_FORMAT, TIME_F_FORMAT,
    DATETIME_P_FORMAT, DATE_P_FORMAT, TIME_P_FORMAT,
)

from .base import FileFormat, identity


class JSONFormat(FileFormat):

    SERIALIZERS = {
        'datetime': lambda d: d.strftime(DATETIME_F_FORMAT),
        'date': lambda d: d.strftime(DATE_F_FORMAT),
        'time': lambda d: d.strftime(TIME_F_FORMAT),
        'number': float,
        'duration': lambda d: isodate.duration_isoformat(d),
        'geopoint': lambda d: list(map(float, d)),
        'yearmonth': lambda d: '{:04d}-{:02d}'.format(*d),
    }

    NULL_VALUE = None

    PYTHON_DIALECT = {
        'date': {
            'format': DATE_P_FORMAT
        },
        'time': {
            'format': TIME_P_FORMAT
        },
        'datetime': {
            'format': DATETIME_P_FORMAT
        },
    }

    def __init__(self, file, schema, **options):
        self.initialize_file(file)
        super(JSONFormat, self).__init__(file, schema, default_serializer=identity, **options)

    def initialize_file(self, file):
        file.write('[')
        file.__first = True

    @classmethod
    def prepare_resource(cls, resource):
        descriptor = resource.descriptor
        descriptor['encoding'] = 'utf-8'
        descriptor['path'] = str(Path(descriptor['path']).with_suffix('.json'))
        descriptor['format'] = 'json'
        descriptor['mediatype'] = 'text/json'
        descriptor['profile'] = 'tabular-data-resource'
        super(JSONFormat, cls).prepare_resource(resource)

    def write_transformed_row(self, transformed_row):
        if not self.writer.__first:
            self.writer.write(',')
        else:
            self.writer.__first = False
        self.writer.write(json.dumps(transformed_row, sort_keys=True, ensure_ascii=True))

    def finalize_file(self):
        self.writer.write(']')



from tabulator.parsers.json import JSONParser


class GeoJsonParser(JSONParser):
    options = []

    def __init__(self, loader, force_parse=False):
        super().__init__(loader, force_parse=force_parse, property='features')

    @property
    def extended_rows(self):
        iterator = super().extended_rows
        for row_number, keys, values in iterator:
            row = dict(zip(keys, values))
            properties = row.get('properties', dict())
            properties['__geometry'] = row.get('geometry')
            items = list(properties.items())
            yield row_number, list(x[0] for x in items), list(x[1] for x in items)



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import io
from setuptools import setup, find_packages


# Helpers
def read(*paths):
    """Read a text file."""
    basedir = os.path.dirname(__file__)
    fullpath = os.path.join(basedir, *paths)
    contents = io.open(fullpath, encoding='utf-8').read().strip()
    return contents


# Prepare
PACKAGE = 'dataflows'
NAME = PACKAGE.replace('_', '-')
INSTALL_REQUIRES = [
    'tabulator>=1.53.5',
    'datapackage>=1.5.0',
    'tableschema>=1.20',
    'kvfile>=0.0.9',
    'click',
    'jinja2',
    'awesome-slugify',
    'inquirer',
    'tabulate',
    'tableschema-sql',
    'xmljson',
    'bitstring>=3',
    'python-dateutil',
    'openpyxl',
    'sqlalchemy<2'
]
SPEEDUP_REQUIRES = [
    'plyvel',
]
LINT_REQUIRES = [
    'pylama',
    'pylama_quotes',
    'pyflakes<2.5',
]
TESTS_REQUIRE = [
    'mock',
    'pytest',
    'pytest-cov',
    'coverage',
    'lxml',
]
README = read('README.md')
VERSION = read(PACKAGE, 'VERSION')
PACKAGES = find_packages(exclude=['examples', 'tests', '.tox'])

# Run
setup(
    name=NAME,
    version=VERSION,
    packages=PACKAGES,
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    extras_require={
        'develop': LINT_REQUIRES + TESTS_REQUIRE,
        'speedup': SPEEDUP_REQUIRES,
    },
    zip_safe=False,
    long_description=README,
    long_description_content_type='text/markdown',
    description='A nifty data processing framework, based on data packages',
    author='Adam Kariv',
    author_email='adam.kariv@gmail.com',
    url='https://github.com/datahq/dataflows',
    license='MIT',
    keywords=[
        'data',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    entry_points={
      'console_scripts': [
        'dataflows = dataflows.cli:cli',
      ]
    }
)



import logging
import json
import datetime
from functools import partial


def identity(x):
    return x


def json_dumps(x):
    return json.dumps(x, ensure_ascii=False)


def comma_separated(x):
    if x is None:
        return None
    elif isinstance(x, list):
        if all(isinstance(i, (str, int, float)) for i in x):
            return ', '.join(str(i) for i in x)
    return json_dumps(x)


class FileFormat():

    PYTHON_DIALECT = {}
    NULL_VALUE = None
    SERIALIZERS = {}
    FILE_MODE = 'w+'

    def __init__(self, writer, schema, temporal_format_property=None, default_serializer=str, **kwargs):

        # Set properties
        self.writer = writer
        self.headers = [f.name for f in schema.fields]
        self.fields = dict((f.name, f) for f in schema.fields)
        self.temporal_format_property = temporal_format_property
        self.missing_values = schema.descriptor.get('missingValues', [])

        # Set fields' serializers
        for field in schema.fields:
            serializer = self.SERIALIZERS.get(field.type, default_serializer)
            if self.temporal_format_property:
                if field.type in ['datetime', 'date', 'time']:
                    format = field.descriptor.get(self.temporal_format_property, None)
                    if format:
                        strftime = getattr(datetime, field.type).strftime
                        serializer = partial(strftime, format=format)
            field.descriptor['serializer'] = serializer

    @classmethod
    def prepare_resource(cls, resource):
        for field in resource.descriptor['schema']['fields']:
            field.update(cls.PYTHON_DIALECT.get(field['type'], {}))

    def __transform_row(self, row):
        try:
            return dict((k, self.__transform_value(v, self.fields[k]))
                        for k, v in row.items())
        except Exception:
            logging.exception('Failed to transform row %r', row)
            raise

    def __transform_value(self, value, field):
        if value is None:
            return self.NULL_VALUE
        # It supports a `tableschema`'s mode of perserving missing values
        # https://github.com/frictionlessdata/tableschema-py#experimental
        if value in self.missing_values:
            return value
        return field.descriptor['serializer'](value)

    def write_transformed_row(self, *_):
        raise NotImplementedError()

    def write_row(self, row):
        transformed_row = self.__transform_row(row)
        self.write_transformed_row(transformed_row)



import os
import itertools
from dataflows import Flow
from .stream import stream
from .unstream import unstream


def _notify_checkpoint_saved(checkpoint_name):

    def step(package):
        yield package.pkg
        for rows in package:
            yield (row for row in rows)
        print(f'checkpoint saved: {checkpoint_name}')

    return step


class checkpoint(Flow):

    def __init__(self, checkpoint_name, checkpoint_path='.checkpoints', steps=None, resources=None):
        if not steps:
            steps = []
        super().__init__(*steps)
        self.checkpoint_name = checkpoint_name
        self.checkpoint_path = os.path.join(checkpoint_path, checkpoint_name)
        self.resources = resources

    @property
    def filename(self):
        return os.path.join(self.checkpoint_path, 'stream.ndjson')

    def exists(self):
        return os.path.exists(self.filename)

    def _preprocess_chain(self):
        if os.path.exists(self.filename):
            print('using checkpoint data from {}'.format(self.checkpoint_path))
            return unstream(self.filename),
        else:
            print('saving checkpoint to: {}'.format(self.checkpoint_path))
            return itertools.chain(self.chain, (stream(self.filename),
                                                _notify_checkpoint_saved(self.checkpoint_name)))

    def handle_flow_checkpoint(self, parent_chain):
        self.chain = itertools.chain(self.chain, parent_chain)
        return [self]



import copy


def update_package(**metadata):

    metadata = copy.deepcopy(metadata)
    if 'resources' in metadata:
        del metadata['resources']

    def func(package):
        package.pkg.descriptor.update(metadata)
        yield package.pkg
        yield from package

    return func


add_metadata = update_package



# -*- coding: utf-8 -*-

from sqlalchemy import sql, create_engine
from tabulator import exceptions
from tabulator.parser import Parser


class ExtendedSQLParser(Parser):
    """Parser to get data from SQL database.
    """

    # Public

    options = [
        'table',
        'order_by',
        'query'
    ]

    def __init__(self, loader, force_parse=False, table=None, order_by=None, query=None):
        if query is None and table is None:
            raise exceptions.TabulatorException('Format `sql` requires `table` or `query` options.')

        # Set attributes
        self.__loader = loader
        self.__table = table
        self.__order_by = order_by
        self.__query = query
        self.__force_parse = force_parse
        self.__engine = None
        self.__extended_rows = None
        self.__encoding = None

    @property
    def closed(self):
        return self.__engine is None

    def open(self, source, encoding=None):
        self.close()
        self.__engine = create_engine(source)
        self.__engine.update_execution_options(stream_results=True)
        self.__encoding = encoding
        self.reset()

    def close(self):
        if not self.closed:
            self.__engine.dispose()
            self.__engine = None

    def reset(self):
        self.__extended_rows = self.__iter_extended_rows()

    @property
    def encoding(self):
        return self.__encoding

    @property
    def extended_rows(self):
        return self.__extended_rows

    # Private
    def __iter_extended_rows(self):
        if self.__query is not None:
            query = self.__query
        else:
            table = sql.table(self.__table)
            order = sql.text(self.__order_by) if self.__order_by else None
            query = sql.select(['*']).select_from(table).order_by(order)
        result = self.__engine.execute(query)
        for row_number, row in enumerate(iter(result), start=1):
            yield (row_number, list(row.keys()), list(row))



from inspect import isfunction

from .. import DataStreamProcessor, schema_validator, ResourceWrapper
from ..base.schema_validator import raise_exception, wrap_handler
from ..helpers import ResourceMatcher


class validate(DataStreamProcessor):

    def __init__(self, *args, resources=None, on_error=None):
        super(validate, self).__init__()
        if on_error is None:
            on_error = raise_exception
        self.on_error = wrap_handler(on_error)
        self.resources = resources
        if len(args) == 2:
            field, validator = args
            assert isinstance(field, str), 'Field name must be a string'
            assert isfunction(validator), 'Validator must be callable'
            validator = self.row_validator(field, validator)
            validator = self.rows_validator(validator)
        elif len(args) == 1:
            validator = args[0]
            assert isfunction(validator), 'Validator must be callable'
            validator = self.rows_validator(validator)
        elif len(args) == 0:
            validator = self.validate_with_schema()
        else:
            assert False, 'Unexpected number of arguments'
        assert validator is not None
        self.validator = validator

    def row_validator(self, field, field_validator):
        def func(row):
            return field_validator(row.get(field))
        return func

    def rows_validator(self, row_validator):
        def func(rows: ResourceWrapper):
            res_name = rows.res.name
            for i, row in enumerate(rows):
                if not row_validator(row):
                    if not self.on_error(res_name, row, i, None, None):
                        continue
                yield row
        return func

    def validate_with_schema(self):
        def func(res):
            yield from schema_validator(res.res, res, on_error=self.on_error)
        return func

    def process_resource(self, res):
        if self.resources.match(res.res.name):
            yield from self.validator(res)
        else:
            yield from super().process_resource()

    def process_datapackage(self, dp):
        self.resources = ResourceMatcher(self.resources, dp)
        return super().process_datapackage(dp)



from datapackage import Package

from .. import DataStreamProcessor, Flow, DataStream


class sources(DataStreamProcessor):

    def __init__(self, *sources):
        super().__init__()
        self.sources: DataStream = [
            Flow(s).datastream()
            for s in sources
        ]

    def process_resources(self, resources):
        yield from super().process_resources(resources)
        source: DataStream
        for source in self.sources:
            for res in source.res_iter:
                yield res

    def process_datapackage(self, dp: Package):
        super().process_datapackage(dp)
        descriptor = dp.descriptor
        source: DataStream
        for source in self.sources:
            res1 = descriptor.pop('resources', [])
            res2 = source.dp.descriptor['resources']
            descriptor.update(source.dp.descriptor)
            descriptor['resources'] = res1 + res2
        dp.commit()
        return dp
